<?php

/**
 * @file
 * Wrapper class for the OCC Hybris API Calls.
 *
 * @see hybris_api_get_client()
 */

class HybrisApi {
  public $last_response;

  private $params = array();

  private $account_service = NULL;
  private $cart_service = NULL;
  private $product_service = NULL;
  private $cart = NULL;
  private $drupal_cart = array();

  private $storage = 'COOKIE';

  static $memoize = array();

  const HYBRIS_SERVER_ERROR = 'HybrisServerError';

  const STEP_CART = 'step_cart';
  const STEP_SHIPPING = 'step_shipping';
  const STEP_BILLING = 'step_billing';

  const HYBRIS_TOKEN = 'hybris_token';
  const HYBRIS_EXPIRES = 'hybris_expires';
  const REFRESH_TOKEN = 'refresh_token';
  const CART_TOKEN = 'cart_token';
  const CART_GUID = 'cart_guid';
  const CART_COUNT = 'cart_count';
  const GIFT_WRAP_SKU = 'GIFT_WRAP';
  const ORDER_ID = 'order_id';
  const PAYMENT_TOKEN = 'payment_token';
  const TRANSACTION_TOKEN = 'transaction_token';
  const CARD_INFO = 'card_info';
  const CARD_TYPE = 'card_type';
  const CART_DATA = 'cart_data';

  const INVALID_TOKEN_ERROR = 'InvalidTokenError';

  const ADDRESS_TYPE_SHIPPING = 'SHIPPING';
  const ADDRESS_TYPE_BILLING = 'BILLING';

  const ORDER_STATUS_CREATED = 'CREATED';
  const ORDER_STATUS_CANCELLED = 'CANCELLED';
  const ORDER_STATUS_SHIPPED = 'SHIPPED';
  const ORDER_STATUS_PENDING = 'PENDING';
  const ORDER_STATUS_PROCESSING = 'PROCESSING';

  /**
   * Constructor for the hybris api instance
   *
   * @access public
   *
   * @param array $params - Hybris credentials & options.
   * @param $refreshToken - default is TRUE
   *  pass FALSE for non-Hybris tasks (e.g. getSessionItem) to avoid
   *  race condition.
   *
   * @return void
   */
  public function __construct($params, $refreshToken = TRUE) {
    $this->params = $params;

    // Check to see if there's an active session.
    if (version_compare(phpversion(), '5.4.0', '>=')) {
      $activeSession = session_status() === PHP_SESSION_ACTIVE ? TRUE : FALSE;
    } else {
      $activeSession = session_id() !== '' ? TRUE : FALSE;
    }

    if ($activeSession) {
      $this->storage = 'SESSION';
      $this->migrateCookies();
    }

    // IMPORTANT: Renewing Hybris Access Token should not be invoked in
    // parallel by multiple ajax processes as they will cause race condition;
    // last renew call will invalidate all previously renewed tokens.
    if ($refreshToken) {
      $this->refreshToken();
    }
  }

  /**
   * Refresh a token if it's there and needs to be refreshed.
   *
   * @access private
   *
   * @return void
   */
  private function refreshToken($force = FALSE) {
    global $user;

    // Not applicable to anonymous users here.
    if (user_is_anonymous()) {
      $this->removeSessionItem(self::HYBRIS_TOKEN);
      $this->removeSessionItem(self::HYBRIS_EXPIRES);
      $this->removeSessionItem(self::REFRESH_TOKEN);
      return;
    }

    $hybris_expires = $this->getSessionItem(self::HYBRIS_EXPIRES);

    // Check to see if current token has expired.
    // Note: subtract 30 seconds to cover any time lag between
    // Hybris and Drupal, or long script execution time.
    if (!$force && !empty($hybris_expires) && $hybris_expires - 30 > time()) {
      return;
    }

    $refresh_token = $this->getSessionItem(self::REFRESH_TOKEN);
    $error = FALSE;

    if (empty($refresh_token)) {
      $error = 'Refresh Token is missing';
    } else {
      $account_service = $this->getAccountService();
      $resp = $account_service->refreshAccessToken($refresh_token);

      if ($resp->code == 200) {
        $this->setSessionItem(self::HYBRIS_TOKEN, $resp->result->access_token);
        $this->setSessionItem(self::REFRESH_TOKEN, $resp->result->refresh_token);
        $this->setSessionItem(self::HYBRIS_EXPIRES, time()+$resp->result->expires_in);
      } else {
        $error = $resp->result->errors;
      }
    }

    // Refresh Token is gone or there's error when trying to renew it.
    // Log user out and send watchdog error.
    if (!empty($error) && !user_access('bypass hybris registration')) {
      session_destroy();

      watchdog(
        'hybris_api',
        '[CRIT] Failed to refresh Hybris token: @error',
        array('@error' => json_encode($error)),
        WATCHDOG_CRITICAL
      );
    }
  }

  /**
   * Create and store the account service connector.
   *
   * @access private
   *
   * @return OCCAccount instance
   */
  private function getAccountService() {
    if (empty($this->account_service)) {
      $this->account_service = new OCCAccount(
          $this->params['url'],
          $this->params['site_id'],
          $this->params['client_id'],
          $this->params['secret'],
          $this->params['accept_language'],
          $this->params['catalog_id'],
          $this->params['user_language'],
          $this->params['currency'],
          $this->params['point_of_service'],
          $this->params['debug']
        );
    }
    return $this->account_service;
  }

  /**
   * Create and store the cart service connector.
   *
   * @access private
   *
   * @return OCCCart instance
   */
  private function getCartService() {
    if (empty($this->cart_service)) {
      $this->cart_service = new OCCCart(
          $this->params['url'],
          $this->params['site_id'],
          $this->params['client_id'],
          $this->params['secret'],
          $this->params['accept_language'],
          $this->params['catalog_id'],
          $this->params['user_language'],
          $this->params['currency'],
          $this->params['point_of_service'],
          $this->params['debug']
        );
    }
    return $this->cart_service;
  }

  private function getProductService() {
    if (empty($this->product_service)) {
      $this->product_service = new OCCProduct(
        $this->params['url'],
        $this->params['site_id'],
        $this->params['client_id'],
        $this->params['secret'],
        $this->params['accept_language'],
        $this->params['catalog_id'],
        $this->params['user_language'],
        $this->params['currency'],
        $this->params['point_of_service'],
        $this->params['debug']
      );
    }

    return $this->product_service;
  }

  /**
   * A cleanup function to handle the change from anonymous user to
   * authenticate user to remove stale anonymous cart cookies.
   *
   * @access private
   *
   * @return void
   */
  private function migrateCookies() {
    $prefix = 'Drupal_visitor_hybris_api_';
    foreach ($_COOKIE as $key => $value) {
      if ($key == $prefix . self::CART_COUNT || strpos($key, $prefix) === FALSE) {
        continue;
      }
      $key = str_replace($prefix, '', $key);
      $this->setSessionItem($key, json_decode($value, TRUE));
      user_cookie_delete(sprintf('%s.%s', 'hybris.api', $key));
      unset($_COOKIE[$prefix.$key]);
    }
  }

  /**
   * Store the user specific data in either cookies or session.
   *
   * @access public
   *
   *
   * @param string $key
   * @param string $value
   *
   * @return void
   */
  public function setSessionItem($key, $value) {
    $value = json_encode($value);
    if ($key != self::CART_COUNT && $this->storage == 'SESSION') {
      $_SESSION['hybris_api'][$key] = $value;
    } else {
      user_cookie_save(array("hybris.api.$key" => $value));
      $_COOKIE["Drupal_visitor_hybris_api_$key"] = $value;
    }
  }

  /**
   * Get the user specific data from either cookies or session.
   *
   * @access private
   *
   * @param string $key
   *
   * @return string
   */
  public function getSessionItem($key) {
    if ($key != self::CART_COUNT && $this->storage == 'SESSION') {
      if (isset($_SESSION['hybris_api'][$key])) {
        return json_decode($_SESSION['hybris_api'][$key], TRUE);
      }
      return NULL;
    }
    if (isset($_COOKIE["Drupal_visitor_hybris_api_$key"])) {
      return json_decode($_COOKIE["Drupal_visitor_hybris_api_$key"], TRUE);
    }
    return NULL;
  }

  /**
   * Remove the user specific data from either cookies or session.
   *
   * @access private
   *
   * @param string $key
   *
   * @return void
   */
  public function removeSessionItem($key) {
    if (!empty($_SESSION['hybris_api'])) {
      unset($_SESSION['hybris_api'][$key]);
    }
    user_cookie_delete("hybris.api.$key");
  }

  /**
   * Get cart token from session. If cart token is empty, initialize cart.
   *
   * @access private
   *
   * @return string
   */
  private function getCartToken() {
    if (!($this->getSessionItem(self::CART_TOKEN))) {
      $this->getCart();
    }
    return $this->getSessionItem(self::CART_TOKEN);
  }

  /**
   * Get user profile details from hybris.
   *
   * @access public
   *
   * @param string $email
   *
   * @return string - customer id
   */
  public function getProfile($email) {
    $account_service = $this->getAccountService();
    $resp = $account_service->getUserProfile($email);

    if ($resp->code != '200') {
      return FALSE;
    }

    if (empty($resp->result)) {
      return FALSE;
    }

    $resp->result->created = strtotime($resp->result->registrationDate);

    return $resp->result;
  }

  /**
   * Authenticate user against hybris
   *
   * @access public
   *
   * @param string $username
   * @param string $password
   *
   * @return string - customer id
   */
  public function authenticateUser($username, $password) {
    $account_service = $this->getAccountService();
    $resp = $account_service->login($username, $password);
    $this->last_response = $resp;

    if ($resp->code != '200') {
      return FALSE;
    }

    if (empty($resp->auth) || empty($resp->auth->result) || empty($resp->auth->result->access_token)) {
      return FALSE;
    }

    $this->setSessionItem(self::HYBRIS_TOKEN, $resp->auth->result->access_token);
    $this->setSessionItem(self::REFRESH_TOKEN, $resp->auth->result->refresh_token);
    $this->setSessionItem(self::HYBRIS_EXPIRES, time()+$resp->auth->result->expires_in);

    $resp->profile->result->created = strtotime($resp->profile->result->registrationDate);

    return $resp->profile->result;
  }
  /**
   * Registers a user against hybris.
   *
   * @access public
   *
   * @param string $username
   *   Desired Username
   * @param string $password
   *   Desired Password
   * @param array $profile
   *   Structured array containing first_name, last_name, newsletter
   */
  public function registerUser($email, $password, array $profile) {
    $account_service = $this->getAccountService();

    $resp = $account_service->register(
        $profile['first_name'],
        $profile['last_name'],
        $email,
        $profile['country'],
        $password,
        $profile['receive_newsletter']
      );

    $this->last_response = $resp;

    if (!empty($resp->result->errors)) {
      return FALSE;
    }

    $this->setSessionItem(self::HYBRIS_TOKEN, $resp->auth->result->access_token);
    $this->setSessionItem(self::REFRESH_TOKEN, $resp->auth->result->refresh_token);
    $this->setSessionItem(self::HYBRIS_EXPIRES, time()+$resp->auth->result->expires_in);

    return $resp;
  }

  /**
   * Update account password
   *
   * @access public
   *
   * @param string $old_pass
   * @param string $new_pass
   *
   * @return bool
   */
  public function updatePassword($old_pass, $new_pass) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $resp = $account_service->updatePassword($token, $old_pass, $new_pass);

    if (is_object($resp)) {
      if (!empty($resp->result->errors)) {
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * Update account email address
   *
   * @access public
   *
   * @param string $email
   * @param string $pass
   *
   * @return bool
   */
  public function updateEmail($email, $pass) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $resp = $account_service->updateEmail($token, $email, $pass);

    if (is_object($resp)) {
      if (!empty($resp->result->errors)) {
        return FALSE;
      }
    }

    // Re-auth to not use the old uid.
    $this->authenticateUser($email, $pass);
    return TRUE;
  }

  /**
   * Force reset a password
   *
   * @access public
   *
   * @param string $email
   * @param string $pass
   *
   * @return bool
   */
  public function resetPassword($email, $pass) {
    $account_service = $this->getAccountService();
    $resp = $account_service->changeUserPassword($email, $pass);

    if (is_int($resp->code) && in_array($resp->code, array(200, 201, 202))) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Return an instance of the cart from the api
   *
   * @access public
   *
   * @param bool $no_cache
   *   (optional) A boolean flag to force reloading cart from Hybris
   *   even if it's available in cache. Default is FALSE.
   *
   * @return $cart
   *   Std Object instance of cart from api
   */
  public function getCart($no_cache = FALSE) {
    global $user;

    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    $cart_guid = $this->getSessionItem(self::CART_GUID);
    $cart_token = $this->getSessionItem(self::CART_TOKEN);

    // Look in cache first if not explicitly called with no-cache.
    if (!$no_cache) {
      if (!empty($this->cart)) {
        return $this->cart;
      }

      $cache_key = $this->createCacheKey('cart_' . $user->uid . '_' . (user_is_anonymous() ? $cart_guid : $cart_token));
      $cache = cache_get($cache_key);
      if (!empty($cache->data)) {
        $this->cart = $cache->data;
        return $this->cart;
      }
    }

    $cart_service = $this->getCartService();

    $this->cart = NULL;
    $error = NULL;

    // Anonymous user.
    if (user_is_anonymous()) {
      if (!empty($cart_guid) && $cart_token == $cart_guid) {
        // Has anonymous cart GUID, get the cart...
        $cartResp = $cart_service->getCart(NULL, $cart_guid);

        // ...and cart is valid
        if (empty($cartResp->result->errors)) {
          $this->cart = $cartResp->result;
        }
      }

      // Either anonymous cart GUID doesn't exist or exists but getCart() fails.
      // Then create a new anonymous cart.
      if (empty($this->cart)) {
        $cartResp = $cart_service->createCart();

        if (!empty($cartResp->result->errors)) {
          $error = $cartResp->result->errors;
        } else {
          $this->cart = $cartResp->result;
        }
      }

      // There should be anonymous cart here.
      if (!empty($this->cart)) {
        $this->setSessionItem(self::CART_TOKEN, $this->cart->guid);
        $this->setSessionItem(self::CART_GUID, $this->cart->guid);
      }
    }
    // Authenticated user.
    else {
      // There's cart token and it's an authenticated cart's token.
      if (!empty($cart_token) && $cart_token != $cart_guid) {
        $cartResp = $cart_service->getCart($user_token, $cart_token);
        if (empty($cartResp->result->errors)) {
          $this->cart = $cartResp->result;
        } else {
          $error = $cartResp->result->errors;
        }
      }

      // Here, cart token is either empty or invalid. Get all carts and
      // pick the first. Or create new if none is available.
      if (empty($this->cart)) {
        // First, get user's carts to see if there are any.
        $cartResp = $cart_service->getCart($user_token);
        if (empty($cartResp->result->errors)) {
          // User has no carts, create a new one.
          if (empty($cartResp->result->carts)) {
            // There's anonymous cart previously, then create new cart and copy
            // the anonymous cart to it. Hybris also deletes the anonymous cart
            // if successful.
            if (!empty($cart_guid)) {
              $cartResp = $this->createCartWithMerge($user_token, $cart_guid);

              // New cart from anonymous cart is created.
              if (empty($cartResp->result->errors)) {
                $this->cart = $cartResp->result;
              }
            }

            // The above fails to create new cart, likely because $cart_guid is
            // invalid. Just create a new cart without anonymous cart copying.
            if (empty($this->cart)) {
              $cartResp = $cart_service->createCart($user_token);
              if (empty($cartResp->result->errors)) {
                $this->cart = $cartResp->result;
              } else {
                $error = $cartResp->result->errors;
              }
            }
          } else {
            // User has carts, pick the first one.
            $this->cart = $cartResp->result->carts[0];
            // Merge in anonymous cart if one exists and with items.
            if (!empty($cart_guid) && $this->getCartCount()) {
              $old_cart_code = $this->cart->code;

              $cartResp = $this->createCartWithMerge($user_token, $cart_guid);
              if (empty($cartResp->result->errors)) {
                $this->cart = $cartResp->result;

                // Delete old cart.
                $cart_service->deleteCart($old_cart_code, $user_token);
              }
            }

          }
        } else {
          // Likely that the user token is invalid (e.g. expired)
          $error = $cartResp->result->errors;
        }
      }

      // Authenticated cart is retrieved or created.
      if (!empty($this->cart)) {
        $this->setSessionItem(self::CART_TOKEN, $this->cart->code);
      }

      // Remove anonymous cart guid because guid is not applicable to
      // authenticated checkout.
      $this->removeSessionItem(self::CART_GUID);
    }

    if (empty($this->cart)) {
      // Show error if Hybris server is unreachable.
      if (!empty($error) && $error[0]->type == self::HYBRIS_SERVER_ERROR) {
        drupal_set_message(self::HYBRIS_SERVER_ERROR, 'warning', FALSE);
      }

      $this->setSessionItem(self::CART_COUNT, 0);

      // Might happen when 2+ (ajax) processes doing Hybris token refresh
      // at the same time causing a race condition. In which case, expire
      // the Hybris access token so the next process can refresh.
      if (!empty($error[0]->type) && $error[0]->type == self::INVALID_TOKEN_ERROR) {
        $this->removeSessionItem(self::HYBRIS_EXPIRES);
      }

      // Critical error, usually server issue.
      watchdog(
        'hybris_api',
        'getCart failed: user=@user; cartCode=@cartCode; cartGUID=@cartGUID, error=@error',
        array('@user' => $user->uid, '@cartCode' => $cart_token, '@cartGUID' => $cart_guid, '@error' => @json_encode($error)),
        WATCHDOG_CRITICAL
      );
    } else {
      // Remove unused cart data that's not applicable to front-end checkout.
      unset($this->cart->deliveryOrderGroups);
    }

    // Cache the Hybris cart.
    if (!empty($this->cart)) {
      $cache_key = $this->createCacheKey('cart_' . $user->uid . '_' . (user_is_anonymous() ? $this->cart->guid : $this->cart->code));
      cache_set($cache_key, $this->cart, 'cache', time() + variable_get('hybris_cart_cache_ttl', 60*60));
    }

    return $this->cart;
  }

  /**
   * Add an sku item to the cart.
   *
   * @access public
   *
   * @param string $sku
   * @param integer $qty
   *
   * @return stdObject response object or FALSE
   */
  public function addCartItem($sku, $qty = 1) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->addItem($this->getCartToken(), $sku, $qty, $user_token);

    if (!empty($resp->result->errors)) {
      watchdog(
        'hybris_api',
        'Adding item to cart failed: @errors',
        array('@errors' => json_encode($resp->result->errors)),
        WATCHDOG_WARNING
      );
    }

    return $resp;
  }

  /**
   * Update the quantity of the cart item.
   *
   * @access public
   *
   * @params string $entryNumber
   * @params integer $qty
   *
   * @return stdObject response object or FALSE
   */
  public function updateCartItem($entryNumber, $qty) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->updateItem($this->getCartToken(), $entryNumber, $qty, $user_token);
    if ($resp->code == 200) {
      return $resp;
    }

    return FALSE;
  }

  /**
   * Remove an item from the cart.
   *
   * @access public
   *
   * @param string $entryNumber
   *
   * @return stdObject response object
   */
  public function removeCartItem($entryNumber) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->deleteItem($this->getCartToken(), $entryNumber, $user_token);

    return $resp;
  }

  /**
   * Get countries to which we can deliver.
   *
   * @access public
   *
   * @return array
   */
  public function getDeliveryCountries() {
    $cache_key = $this->createCacheKey('delivery_countries');
    $result = array();

    // Look in cache first.
    $cache = cache_get($cache_key);
    if (!empty($cache->data)) {
      $result = $cache->data;
    }

    // Hit Hybris if not in cache.
    if (empty($result)) {
      $cart_service = $this->getCartService();
      $resp = $cart_service->getDeliveryCountries($this->params['accept_language']);

      if ($resp->code == 200) {
        $result = $resp->result->countries;
        cache_set($cache_key, $result, 'cache', time() + variable_get('hybris_api_cache_ttl', 60*15));
      }
    }

    return $result;
  }


  /**
   * Get countries which we can bill.
   *
   * @access public
   *
   * @return array
   */
  public function getBillingCountries() {
    $cache_key = $this->createCacheKey('billing_countries');
    $result = array();

    // Look in cache first.
    $cache = cache_get($cache_key);
    if (!empty($cache->data)) {
      $result = $cache->data;
    }

    // Hit Hybris if not in cache.
    if (empty($result)) {
      $cart_service = $this->getCartService();
      $resp = $cart_service->getBillingCountries();

      if ($resp->code == 200) {
        $result = $resp->result->countries;
        cache_set($cache_key, $result, 'cache', time() + variable_get('hybris_api_cache_ttl', 60*15));
      }
    }

    return $result;
  }

  /**
   * Sets shipping address or billing address to cart.
   *
   * @access public
   *
   * @param string $type - either 'BILLING' or 'SHIPPING'
   * @param array $address
   *
   * @return TRUE or errors array
   */
  public function setCartAddress($type, $address) {
    global $user;

    if ($type != self::ADDRESS_TYPE_SHIPPING && $type != self::ADDRESS_TYPE_BILLING) {
      return array(
        'type' => 'InvalidParameterError',
        'message' => 'Type must be either BILLING or SHIPPING',
      );
    }

    $address['type'] = $type;

    if (!empty($address['id'])) {
      unset($address['remarks']);
    }

    $account_service = $this->getAccountService();
    $cart_service = $this->getCartService();

    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    $cart_token = $this->getCartToken();

    $create_addr_function = $type == self::ADDRESS_TYPE_SHIPPING ? 'createDeliveryAddress' : 'createBillingAddress';
    $set_addr_function = $type == self::ADDRESS_TYPE_SHIPPING ? 'setDeliveryAddress' : 'setBillingAddress';

    if (user_is_logged_in()) {
      // If address is a user's Saved Address, update it then set to cart.
      if (!empty($address['id'])) {
        $account_service->updateAddress($user_token, $address['id'], $address);

        // Set the address to cart if it's not already is in cart.
        if ($address['id'] != $address['cart_address_id']) {
          $resp = $cart_service->$set_addr_function($cart_token, $address['id'], $user_token);
        }
      } else {
        // Address is not a Saved Address, set it directly to cart.
        $address['visibleInAddressBook'] = !empty($address['save']) ? 'true' : 'false';

        $resp = $cart_service->$create_addr_function($user->mail, $cart_token, $address);
      }
    } else {
      // Anonymous user.
      $resp = $cart_service->$create_addr_function('anonymous', $cart_token, $address);
    }

    if (empty($resp->result->errors)) {
      return TRUE;
    }

    return $resp->result->errors;
  }

  /**
   * Add an address to the account.
   *
   * @access public
   * @param $address
   * @return bool
   */
  public function addAddress($address) {
    $account_service = $this->getAccountService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    // Not applicable to anonymous user.
    if (empty($user_token)) {
      return FALSE;
    }

    $resp = $account_service->addAddress($user_token, $address);

    if ($resp->code == 201) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Get a delivery modes.
   *
   * @access public
   *
   * @return stdObject response object
   */
  public function getDeliveryModes() {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->getAvailableDeliveryModes($this->getCartToken(), $user_token);

    if ($resp->code == 200) {
      if (!empty($resp->result->deliveryModes)) {
        return $resp->result->deliveryModes;
      }
      return array();
    }

    return array();
  }

  /**
   * Set a delivery mode.
   *
   * @access public
   *
   * @return stdObject response object
   */
  public function setDeliveryMode($delivery_mode) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->setDeliveryMode($this->getCartToken(), $delivery_mode, $user_token);

    if ($resp->code == 200) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Set a preview delivery mode.
   *
   * @access public
   *
   * @return stdObject response object
   */
  public function setPreviewDeliveryMode($delivery_mode) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->setPreviewDeliveryMode($this->getCartToken(), $delivery_mode, $user_token);

    if ($resp->code == 200) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Helper method to get user preferences out of Hybris.
   *
   * @access public
   *
   * @return bool
   * @throws Exception
   */
  public function getUserPrefs() {
    $account_service = $this->getAccountService();

    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($user_token)) {
      return FALSE;
    }

    $resp = $account_service->getPrefs($user_token);

    if ($resp->code != 200) {
      throw new \Exception('Hybris Error.');
    }

    if (!empty($resp->result->errors)) {
      return FALSE;
    }

    return $resp->result;
  }

  /**
   * Get address details from hybris.
   *
   * @access public
   *
   * @param string $id
   *
   * @return mixed
   */
  public function getAccountAddresses($id = NULL) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $resp = $account_service->getAddress($token, $id);

    if ($resp->code == 200) {
      if (!empty($resp->result->address)) {
        return $resp->result->address;
      }
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Add address to account in hybris.
   *
   * @access public
   *
   * @param array $address
   *
   * @return mixed
   */
  public function addAccountAddress($address) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $resp = $account_service->addAddress($token, $address);

    if ($resp->code == 201) {
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Update address to account in hybris.
   *
   * @access public
   *
   * @param string $address_id
   * @param array $address
   *
   * @return mixed
   */
  public function updateAccountAddress($address_id, $address) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $address['town'] = $address['city'];

    $resp = $account_service->updateAddress($token, $address_id, $address);

    if ($resp->code == 200) {
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Remove address to account in hybris.
   *
   * @access public
   *
   * @param string $address_id
   *
   * @return mixed
   */
  public function removeAccountAddress($address_id) {
    $account_service = $this->getAccountService();

    $token = $this->getSessionItem(self::HYBRIS_TOKEN);
    if (empty($token)) {
      return FALSE;
    }

    $resp = $account_service->deleteAddress($token, $address_id);

    if ($resp->code == 200) {
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Used to get product attributes, can pass $since for updated products.
   *
   * @param NULL $since
   *  Starting timestamp.
   *
   * @return stdClass or FALSE
   */
  public function exportProducts($since = NULL, $params = array()) {
    $default_params = array(
      'catalog' => variable_get('hybris_api_client_catalog_id'),
      'version' => 'online',
      'pageSize' => 10,
    );

    $params = array_merge($default_params, $params);

    if (!empty($since)) {
      $params['timestamp'] = date('c', $since);
    }

    $resp = $this->getProductService()->exportProducts($params);

    if ($resp->code == 200) {
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Wrapper to stash the payment token as it doesn't get attached to the cart
   * until the moment the cart is submitted.
   *
   * @access public
   *
   * @param string $payment_type
   *   e.g. creditcard, paypal, debitcard
   * @param string $payment_token
   *   PayNext payment token
   * @param string $card_type
   *   Credit card type, Visa, Mastercard, etc.
   *
   * @return TRUE or Hybris errors array
   */
  public function setPayment($payment_type, $payment_token, $card_type) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    $cart_token = $this->getCartToken();

    $resp = $cart_service->setPaymentMode($cart_token, $payment_type, $user_token);
    if (empty($resp->result->errors)) {
      $this->setSessionItem(self::PAYMENT_TOKEN, $payment_token);
      $this->setSessionItem(self::CARD_TYPE, $card_type);
      return TRUE;
    }

    return $resp->result->errors;
  }

  /**
   * Get the stashed payment token.
   *
   * @access public
   *
   * @return string payment_token
   */
  public function getPaymentToken() {
    return $this->getSessionItem(self::PAYMENT_TOKEN);
  }

  /**
   * Get cart's shipping address.
   *
   * @access public
   *
   * @return stdObject shipping address
   */
  public function getShippingAddress() {
    return !empty($this->cart->deliveryAddress) ? $this->cart->deliveryAddress : NULL;
  }

  /**
   * Finalize and create the order against Hybris.
   *
   * @access public
   *
   * @param string $transaction_id
   *   A transaction token obtained from PayNext payment authorization.
   *
   * @return Hybris response object
   */
  public function createOrder($transaction_id) {
    $cart_service = $this->getCartService();

    $cart = $this->getCart();
    $authorized_amount = $cart->totalPrice->value;

    $cart_token = $this->getCartToken();
    $payment_token = $this->getSessionItem(self::PAYMENT_TOKEN);
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);

    $resp = $cart_service->createOrder(
      $cart_token,
      $authorized_amount,
      $transaction_id,
      $payment_token,
      $user_token
    );

    return $resp;
  }

  /**
   * Clears out the old payment token and order id.
   *
   * @access public
   *
   * @return void
   */
  public function clearOrderSession() {
    global $user;
    $cache_key = $this->createCacheKey('order_' . $user->uid . '_' . $this->getOrderId());
    cache_clear_all($cache_key, 'cache');

    $this->removeSessionItem(self::PAYMENT_TOKEN);
    $this->removeSessionItem(self::ORDER_ID);
    $this->removeSessionItem(self::CARD_TYPE);
    $this->removeSessionItem(self::TRANSACTION_TOKEN);
    $this->removeSessionItem(self::CARD_INFO);
    $this->removeSessionItem(self::CART_DATA);
  }

  /**
   * Clears out the old cart info stored in cookie/session.
   *
   * @access public
   *
   * @return void
   */
  public function clearCartSession() {
    $this->removeSessionItem(self::CART_TOKEN);
    $this->removeSessionItem(self::CART_GUID);
    $this->removeSessionItem(self::CART_COUNT);
  }

  /**
   * Store order id in user's session.
   *
   * @access public
   *
   * @param string $orderId
   *
   * @return stdObject order confirmation
   */
  public function setOrderId($orderId) {
    return $this->setSessionItem(self::ORDER_ID, $orderId);
  }

  /**
   * Get the order id stored in session.
   *
   * @access public
   *
   * @return string
   */
  public function getOrderId() {
    return $this->getSessionItem(self::ORDER_ID);
  }

  /**
   * Get past orders.
   *
   * @access public
   *
   * @return array orders
   */
  public function getPastOrders($count = 10, $page = 0) {
    $cart_service = $this->getCartService();
    $resp = $cart_service->getHistoricalOrders($this->getSessionItem(self::HYBRIS_TOKEN), $count, $page);

    if ($resp->code == 200) {
      return $resp->result;
    }

    return array();
  }

  /**
   * Gets order details from Hybris.
   *
   * @access public
   *
   * @param string $order_id
   * @param bool $no_cache
   *
   * @return object order
   */
  public function getOrderDetail($order_id, $no_cache = FALSE) {
    global $user;
    $cache_key = $this->createCacheKey('order_' . $user->uid . '_' . $order_id);

    if (!$no_cache) {
      $cache = cache_get($cache_key);
      if (!empty($cache->data)) {
        return $cache->data;
      }
    }

    $cart_service = $this->getCartService();
    $resp = $cart_service->getOrderDetails($order_id, $this->getSessionItem(self::HYBRIS_TOKEN));

    if (empty($resp->result->errors)) {
      // Order cancelable logic: within 30 minutes
      $orderTS = date_create($resp->result->created);
      $now = new DateTime('now');
      $diff = ($now->getTimestamp() - $orderTS->getTimestamp()) / 60;
      $resp->result->cancelable = ($resp->result->status == self::ORDER_STATUS_CREATED && $diff <= 30) ? TRUE : FALSE;

      cache_set($cache_key, $resp->result, 'cache', time() + variable_get('hybris_cart_cache_ttl', 60*60));
      return $resp->result;
    }

    return NULL;
  }

  /**
   * Add promo code to cart.
   *
   * @param string $promo_code
   *
   * @return stdObject
   */
  public function addPromoCode($promo_code) {
    $cart_service = $this->getCartService();

    $resp = $cart_service->addVoucher(
      $this->getCartToken(),
      $promo_code,
      $this->getSessionItem(self::HYBRIS_TOKEN)
    );

    if (!empty($resp->result->errors)) {
      return $resp->result->errors;
    }

    return TRUE;
  }

  /**
   * Add Kit
   *
   * @access public
   *
   * @param string $kit_id
   * @param array $skus
   *
   * @return stdClass {code, result}
   */
  public function addKit($kit_id, $skus) {
    $cart_service = $this->getCartService();

    $skus = implode(',', $skus);
    $resp = $cart_service->addKit(
      $this->getCartToken(),
      $kit_id,
      $skus,
      $this->getSessionItem(self::HYBRIS_TOKEN)
    );

    return $resp;
  }

  /**
   * Update Kit
   *
   * @access public
   *
   * @param string $kit_id
   * @param array $skus
   *
   * @return stdClass {code, result}
   */
  public function updateKit($kit_id, $skus) {
    $cart_service = $this->getCartService();

    $skus = implode(',', $skus);
    $resp = $cart_service->updateKit(
      $this->getCartToken(),
      $kit_id,
      $skus,
      $this->getSessionItem(self::HYBRIS_TOKEN)
    );

    return $resp;
  }

  /**
   * Remove Kit
   *
   * @access public
   *
   * @param string $kit_id
   *
   * @return boolean
   */
  public function deleteKit($kit_id) {
    $cart_service = $this->getCartService();

    $resp = $cart_service->deleteKit(
      $this->getCartToken(),
      $kit_id,
      $this->getSessionItem(self::HYBRIS_TOKEN)
    );

    if ($resp->code == 200) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Get session stored country
   *
   * @access public
   *
   * @return array
   */
  public function getDeliveryCountry() {
    return !empty($this->cart->deliveryCountry) ? $this->cart->deliveryCountry : $this->params['default_country'];
  }

  /**
   * Sets delivery country in Hybris cart.
   *
   * @access public
   *
   * @param string $countryIso
   *
   * @return bool
   */
  public function setDeliveryCountry($countryIso) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    $cart_token = $this->getCartToken();

    $resp = $cart_service->setDeliveryCountry($cart_token, $countryIso, $user_token);

    if (!empty($resp->result->errors)) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get session stored mode
   *
   * @access public
   *
   * @return array
   */
  public function getDeliveryMode() {
    return !empty($this->cart->deliveryMode) ? $this->cart->deliveryMode->code : NULL;
  }

  /**
   * Get available delivery modes for a specific ship-to country.
   *
   * @access public
   *
   * @param string $countryIso
   *
   * @return array
   */
  public function getDeliveryModesForCountry($countryIso) {
    $cacheKey = $this->createCacheKey('delivery_modes_' . $countryIso);
    $result = array();

    // Check in buffer.
    if (!empty(self::$memoize['delivery_modes'][$countryIso])) {
      $result = self::$memoize['delivery_modes'][$countryIso];
    }

    // Check in Drupal cache.
    if (empty($result)) {
      $cache = cache_get($cacheKey);
      if (!empty($cache->data)) {
        self::$memoize['delivery_modes'][$countryIso] = $cache->data;
        $result = $cache->data;
      }
    }

    // Not in cache; hit Hybris.
    if (empty($result)) {
      $cart_service = $this->getCartService();
      $resp = $cart_service->getDeliveryModes($countryIso);
      if ($resp->code == 200) {
        foreach ($resp->result->deliveryModes as $mode) {
          $result[$mode->code] = $mode;
        }
        cache_set($cacheKey, $result, 'cache', time() + variable_get('hybris_api_cache_ttl', 60*15));
        self::$memoize['delivery_modes'][$countryIso] = $result;
      }
    }

    // Update shipping options with true costs. True shipping costs come from
    // cart after everything is factored in, e.g. cheaper for large order.
    if (!empty($result) && current_path() != 'system/ajax') {
      $hybris_cart = $this->getCart();
      $drupal_cart = $this->getDrupalCart();
      $free_shipping_promos = $drupal_cart['shipping_promotions']['applied'];
      foreach ($result as $key => &$mode) {
        // If delivery cost is free via promotion, display it as Free.
        if (!empty($free_shipping_promo = $free_shipping_promos[$mode->code])
          && $free_shipping_promo->promotion->actualDeliveryMode->code == $mode->code) {

          $mode->deliveryCostAdjusted = $free_shipping_promo->promotion->actualDeliveryMode->deliveryCost;
          if (empty($mode->deliveryCostAdjusted->value)) {
            $mode->deliveryCostAdjusted->formattedValue = t('Free', array(), array('context' => '0 value'));
          }
        }
        // Use final calculated delivery cost in cart if available.
        elseif (!empty($hybris_cart->deliveryMode) && !empty($hybris_cart->deliveryCost) && $hybris_cart->deliveryMode->code == $mode->code) {
          $mode->deliveryCost = $hybris_cart->deliveryCost;
          if (empty($mode->deliveryCost->value)) {
            $mode->deliveryCost->formattedValue = t('Free', array(), array('context' => '0 value'));
          }
        }
      }
    }

    return $result;
  }

  /**
   * Update user profile in Hybris.
   *
   * @access public
   *
   * @param array $profile
   *  Structured array containing first_name, last_name,
   *  birth_date (mm-dd), country (isocode), newsletter (boolean)
   *
   * @return boolean
   */
  public function updateProfile(array $profile) {
    $account_service = $this->getAccountService();

    $resp = $account_service->updateAccount(
      $this->getSessionItem(self::HYBRIS_TOKEN),
      $profile['first_name'],
      $profile['last_name'],
      $profile['birth_date'],
      $profile['country'],
      $profile['receive_newsletter']
    );

    if ($resp->code == 200) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Replaces one sku for another, and maintains quantity.
   *
   * @access public
   *
   * @param string $current_sku_code
   * @param string $replacement_sku_code
   *
   * @return bool
   */
  public function replaceCartItem($current_sku_code, $replacement_sku_code) {
    $cart = $this->getCart();

    $quantity = 1;
    $entry = NULL;
    foreach ($cart->entries as $entry) {
      if ($entry->product->code == $current_sku_code) {
        $quantity = $entry->quantity;
        $entry = $entry->entryNumber;
        break;
      }
    }

    $this->removeCartItem($entry);
    return $this->addCartItem($replacement_sku_code, $quantity);
  }

  /**
   * Returns kit products to be used in the Value Kit Builder (VKB).
   *
   * @return stdClass
   */
  public function exportKits() {
    $resp = $this->getProductService()->exportKits();

    if ($resp->code == 200) {
      return $resp->result;
    }

    return FALSE;
  }

  /**
   * Returns list of auto-apply promotions.
   *
   * @return stdClass
   */
  public function getAutoApplyPromotions() {
    $cacheKey = $this->createCacheKey('shipping_promotions');
    $cache = cache_get($cacheKey);

    if (!empty($cache->data)) {
      return $cache->data;
    }

    $resp = $this->getCartService()->getAutoApplyPromotions();
    if ($resp->code == 200) {
      cache_set($cacheKey, $resp->result->promotions, 'cache', time() + variable_get('hybris_api_cache_ttl', 60*15));
      return $resp->result->promotions;
    }

    return FALSE;
  }

  /**
   * Get order details using the super system token.
   *
   * @access public
   *
   * @param string $user_id
   * @param string $order_id
   *
   * @return stdObject order
   */
  public function getUserOrderDetail($user_id, $order_id) {
    $cart_service = $this->getCartService();
    $resp = $cart_service->getUserOrderDetails($user_id, $order_id);

    if ($resp->code == 200) {
      return $resp->result;
    }

    return array();
  }

  /**
   * Cancels an order.
   *
   * @access public
   *
   * @param string $user_id
   * @param string $order_id
   * @param string $consignment_id
   *
   * @return boolean
   */
  public function cancelUserOrder($user_id, $order_id, $consignment_id) {
    $cart_service = $this->getCartService();
    $resp = $cart_service->cancelOrder($user_id, $order_id, $consignment_id);

    if ($resp->code == 200) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Sets cart count.
   */
  public function setCartCount($count) {
    $this->setSessionItem(self::CART_COUNT, $count);
  }

  /**
   * Gets cart count.
   */
  public function getCartCount() {
    $cartCount = $this->getSessionItem(self::CART_COUNT);
    return !empty($cartCount) ? $cartCount : 0;
  }

  /**
   * Sets cart's gift wrap and/or gift message
   *
   * @param $giftMessage - optional gift wrap message
   *
   * @return TRUE | FALSE
   */
  public function setGiftWrap($giftMessage = NULL) {
    $cart_service = $this->getCartService();
    $user_token = $this->getSessionItem(self::HYBRIS_TOKEN);
    $cart_token = $this->getCartToken();

    $resp = $cart_service->setGiftWrap($cart_token, $user_token);
    if (empty($resp->result->errors) && !empty($giftMessage)) {
      $resp = $cart_service->addGiftMessage($cart_token, $giftMessage, $user_token);
    }

    if (empty($resp->result->errors)) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Gets cart's gift wrap status.
   *
   * @param $cart - Hybris cart object
   *
   * @return Associative array of:
   *  - status: whether cart has gift wrap item
   *  - message: gift wrap message
   *  - gift_wrap_sku_entry_number: entry # of gift wrap item in cart if exists
   */
  public function getGiftWrapStatus($cart) {
    $result = array(
      'status' => FALSE,
      'message' => NULL,
      'gift_wrap_sku_entry_number' => NULL,
    );

    if (empty($cart)) {
      return $result;
    }

    // Check if cart has GIFT_WRAP sku.
    foreach ($cart->entries as $entry) {
      if (!empty($entry->product->code) && $entry->product->code == self::GIFT_WRAP_SKU) {
        $result['status'] = TRUE;
        $result['gift_wrap_sku_entry_number'] = $entry->entryNumber;
        break;
      }
    }

    if (!empty($result['status'])) {
      $result['message'] = $cart->giftWrapMessage;
    }

    return $result;
  }

  /**
   * Returns Drupal cart.
   */
  public function getDrupalCart() {
    return $this->drupal_cart;
  }

  /**
   * Sets Drupal cart.
   */
  public function setDrupalCart($cart) {
    $this->drupal_cart = $cart;
  }

  /**
   * Get the stashed credit card type.
   *
   * @access public
   *
   * @return string
   */
  public function getCardType() {
    return $this->getSessionItem(self::CARD_TYPE);
  }

  /**
   * Creates an authenticated cart and merge an anonymous cart content to it.
   *
   * @param $userToken
   *   Authenticated user token.
   * @param $anonymousCartGuid
   *   Anonymous cart GUID to merge from.
   * @param $toMergeCartGuid
   *   (optional) Existing authenticated cart GUID to merge into.
   */
  public function createCartWithMerge($userToken, &$anonymousCartGuid, $toMergeCartGuid = NULL) {
    $cart_service = $this->getCartService();

    $resp = $cart_service->createCart($userToken, $anonymousCartGuid, $toMergeCartGuid);

    // Get full cart content after successful merge.
    if (empty($resp->result->errors)) {
      $resp = $cart_service->getCart($userToken, $resp->result->code);
      $anonymousCartGuid = NULL;
    }

    return $resp;
  }

  /**
   * Remove promo code from cart.
   *
   * @param string $promo_code
   *
   * @return stdObject
   */
  public function removePromoCode($promo_code) {
    $cart_service = $this->getCartService();

    $resp = $cart_service->deleteVoucher(
      $this->getCartToken(),
      $promo_code,
      $this->getSessionItem(self::HYBRIS_TOKEN)
    );

    if (!empty($resp->result->errors)) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Creates cache key to cache Hybris data.
   *
   * @param string $key
   *
   * @return string
   */
  public function createCacheKey($key) {
    $env = !empty($_ENV['DRUPAL_ENVIRONMENT']) ? $_ENV['DRUPAL_ENVIRONMENT'] : 'local';
    return sprintf('hybris_api_%s_%s_%s', $this->params['site_id'], $env, $key);
  }

  /**
   * Formats a money amount for display based on the store's currency isocode.
   *
   * @param $amount
   *
   * @return string
   */
  public function formatCurrency($amount) {
    $result = currency_load($this->params['currency'])->format($amount);
    $result = str_replace(' ', '&nbsp;', $result);
    return $result;
  }

  /**
   * Returns the store's currency symbol.
   */
  public function getCurrencySymbol() {
    return currency_load($this->params['currency'])->sign;
  }

  /**
   * Returns the store's default country.
   */
  public function getDefaultCountry() {
    return $this->params['default_country'];
  }

  /**
   * Sets transaction token into session.
   *
   * @access public
   *
   * @param $transaction_token
   */
  public function setTransactionToken($transaction_token) {
    return $this->setSessionItem(self::TRANSACTION_TOKEN, $transaction_token);
  }

  /**
   * Gets transaction token stored in session.
   *
   * @access public
   *
   * @return string
   */
  public function getTransactionToken() {
    return $this->getSessionItem(self::TRANSACTION_TOKEN);
  }

  /**
   * Clears the cached cart.
   *
   */
  public function clear_cached_cart() {
    global $user;
    $cart_guid = $this->getSessionItem(self::CART_GUID);
    $cart_token = $this->getSessionItem(self::CART_TOKEN);

    $cache_key = $this->createCacheKey('cart_' . $user->uid . '_' . (user_is_anonymous() ? $cart_guid : $cart_token));
    cache_clear_all($cache_key, 'cache');
  }

  /**
   * Sets card info in session, only non-sensitive info.
   *
   * @param array $info
   *
   */
  public function setCardInfo($info) {
    return $this->setSessionItem(self::CARD_INFO, $info);
  }

  /**
   * Gets the stashed card info.
   *
   * @access public
   *
   * @return string
   */
  public function getCardInfo() {
    return $this->getSessionItem(self::CARD_INFO);
  }

  /**
   * Checks to see whether an item (SKU) is in cart.
   *
   * @param $sku
   *
   * @return bool
   */
  public function cartHasItem($sku) {
    $this->getCart();

    if (empty($sku) || empty($this->cart->entries)) {
      return FALSE;
    }

    foreach ($this->cart->entries as $entry) {
      if ($entry->product->code == $sku) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Saves custom checkout data.
   *
   * @param array $data
   *
   */
  public function setCustomData(array $data) {
    $existing_data = $this->getSessionItem(self::CART_DATA);
    $data = is_array($existing_data) ? array_merge($existing_data, $data) : $data;
    $this->setSessionItem(self::CART_DATA, $data);
  }

  /**
   * Returns custom checkout data.
   *
   * @return array
   *
   */
  public function getCustomData() {
    return $this->getSessionItem(self::CART_DATA);
  }

  /**
   * Gets product stock level.
   *
   * @param string $sku
   * @param boolean $bypass_cache - default is FALSE
   *
   * @return int or false if the API call fails
   *
   */
  public function getStockLevel($sku, $bypass_cache = FALSE) {
    if (empty($sku)) {
      return FALSE;
    }

    $stock_level = FALSE;
    $cache_key = $this->createCacheKey('stock_level_' . $sku);

    // Look in cache if not explicitly bypassed.
    if (!$bypass_cache && !empty($cache = cache_get($cache_key)) && is_int($cache->data)) {
      $stock_level = $cache->data;
    }

    // Hit the Hybris API if not in cache.
    if (!is_int($stock_level)) {
      $resp = $this->getProductService()->getProductStockLevel($sku);
      if ($resp->code == 200 && !empty($resp->result)) {
        $stock_level = $resp->result->stockLevel;

        // Cache it.
        if (is_int($stock_level)) {
          cache_set($cache_key, $stock_level, 'cache', time() + variable_get('hybris_stock_level_ttl', 60*60));
        }
      }
    }

    return $stock_level;
  }
}
